/* z81/xz81, Linux console and X ZX81/ZX80 emulators.
 * Copyright (C) 1994 Ian Collier. z81 changes (C) 1995-2001 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/select.h>
#include <termios.h>
#include <ctype.h>
#include "common.h"
#include "sound.h"
#include "z80.h"
#include "allmain.h"

extern int scrn_freq;
int hsize=ZX_VID_VGA_WIDTH,vsize=ZX_VID_VGA_HEIGHT;
FILE *sf;

static struct termios st = {0};

void initts()
{
        if (tcgetattr(0, &st) < 0) perror("tcgetattr()");
        st.c_lflag &= ~ICANON;
        st.c_lflag &= ~ECHO;
        st.c_cc[VMIN] = 1;
        st.c_cc[VTIME] = 0;
        if (tcsetattr(0, TCSANOW, &st) < 0) perror("tcsetattr ICANON");
}

void resetts()
{
        st.c_lflag |= ICANON;
        st.c_lflag |= ECHO;
        if (tcsetattr(0, TCSANOW, &st) < 0) perror ("tcsetattr ~ICANON");
}

char getch() {
        char c=0;
	int n;
	struct timeval tv = { 0L, 0L };
	fd_set fds;

	FD_ZERO(&fds);
	FD_SET(0, &fds);

        if (select(1, &fds, NULL, NULL, &tv)>0) {
          n = read(0, &c, 1);
	  if (n<0) perror("read()");
	}

        return c;
}

void exit_program(void)
{
resetts();
#ifdef OSS_SOUND_SUPPORT
sound_end();
#endif
zxpclose();
exit(0);
}

int sixel_output(int n, char c)
{
	int i;

	if (n > 3) {
	  return fprintf(sf,"!%d%c", n, c);
	} else {
	  for (i=0; i<n; i++) fputc(c,sf);
	}
	return n;
}

// set to 1 for double sized screen; unfortunately scrolling is slow

#if 0

void sixel_screen()
{
	unsigned int ptr=0, ofs;
	int i, j1, j2, k, l, s, ps, ns, no;
	unsigned char msk;

	fprintf(sf,"\033Pq#9~!782@~-\n");

	for (j1=0; j1<ZX_VID_FULLHEIGHT/6; j1++) {
	for (j2=0; j2<2; j2++) {
	  fprintf(sf,"~!7?#1");
	  ps = 0;
	  ns = 0;
	  no = 0;
	  for (i=0; i<ZX_VID_FULLWIDTH/8; i++) {
	    msk = 0x80;
	    for (k=0; k<8; k++) {
	      s = 0;
	      ofs = 0;
	      if (j2==0) {
	        for (l=0; l<3; l++) {
	          if (scrnbmp[ptr+i+ofs] & msk) s += (3 << l*2);
		  ofs += ZX_VID_FULLWIDTH/8;
	        }
	      } else {
	        for (l=0; l<3; l++) {
	          if (scrnbmp[ptr+i+ofs+3*(ZX_VID_FULLWIDTH/8)] & msk) s += (3 << l*2);
		  ofs += ZX_VID_FULLWIDTH/8;
	        }
	      }
	      if (invert_screen) s ^= 0x3f;
	      s += 63;
	      ns += 2;
	      if (s!=ps && ps!='\0') {
		no += sixel_output(ns, ps);
		if (no>70) { fputc('\n',sf); no=0; }	        
		ns = 0;
	      }
	      ps = s;
	      msk >>= 1;
	    }
	  }
	  sixel_output(ns, ps);
	  fprintf(sf,"#9!7?~-\n");
	  if (j2==1) ptr += (6*ZX_VID_FULLWIDTH)/8;
	}}

	fprintf(sf,"~!782_~-\n\033\\\n");
}

#else

void sixel_screen()
{
	unsigned int ptr=0, ofs;
	int i, j, k, l, s, ps, ns, no;
	unsigned char msk;

	fprintf(sf,"\033Pq#9~!398@~-\n");

	for (j=0; j<ZX_VID_FULLHEIGHT/6; j++) {
	  fprintf(sf,"~!7?#1");
	  ps = 0;
	  ns = 0;
	  no = 0;
	  for (i=0; i<ZX_VID_FULLWIDTH/8; i++) {
	    msk = 0x80;
	    for (k=0; k<8; k++) {
	      s = 0;
	      ofs = 0;
	      for (l=0; l<6; l++) {
	        if (scrnbmp[ptr+i+ofs] & msk) s += (1 << l);
		ofs += ZX_VID_FULLWIDTH/8;
	      }
	      if (invert_screen) s ^= 0x3f;
	      s += 63;
	      ns++;
	      if (s!=ps && ps!='\0') {
		no += sixel_output(ns, ps);
		if (no>70) { fputc('\n',sf); no=0; }	        
		ns = 0;
	      }
	      ps = s;
	      msk >>= 1;
	    }
	  }
	  sixel_output(ns, ps);
	  fprintf(sf,"#9!7?~-\n");
	  ptr += (6*ZX_VID_FULLWIDTH)/8;
	}

	fprintf(sf,"~!398_~-\n\033\\\n");
}

#endif

/* redraw the screen */
void update_scrn(void)
{
  int x,y;
  unsigned char *ptr,*optr;

for(y=0;y<ZX_VID_VGA_HEIGHT;y++)
  {
  ptr=scrnbmp+(y+ZX_VID_VGA_YOFS)*ZX_VID_FULLWIDTH/8+ZX_VID_VGA_XOFS/8;
  optr=scrnbmp_old+(ptr-scrnbmp);
  for(x=0;x<ZX_VID_VGA_WIDTH;x+=8,ptr++,optr++)
    {
        if(*ptr!=*optr) { refresh_screen=1; goto breakout; }
    }
  }

breakout:
  if (refresh_screen) {
    sixel_screen();
/* now, copy new to old for next time */
    memcpy(scrnbmp_old,scrnbmp,ZX_VID_FULLHEIGHT*ZX_VID_FULLWIDTH/8);
    refresh_screen=0;
  }

}


/* read keyboard and update keyports[] */
void check_events()
{
  static int state=0;
  int b, y;
  char c, cc;
  c = getch();

  if (c=='?') {
    help = !help;
    return;
  }

  if (c!=0) {
    state = 1;
  } else if (state>0) {
    if (state==3) {
      state = 0;
    } else {
      state++;
      return;
    }
  }

  cc = tolower(c);
  if (c=='_') { exit_program(); }
  if (c==8 || c==127) c=')';

for(y=0;y<8;y++)		/* 8 half-rows */
  {
  b=0;	/* we set bits in b as appropriate */
  switch(y)	/* below comments given in order b1->b5 */
    {
    /* left-hand side */
    case 0:	/* sft,z,x,c,v */
      if ((c>='A' && c<='Z') || c=='!' || c=='@' || c=='#' || c=='$' || c=='%' || c=='^' || c=='&' || c=='*' || c=='(' || c==')' || c=='>' || c=='\'' || c=='`') b|=1;
      if(cc=='z') b|=2;
      if(cc=='x') b|=4;
      if(cc=='c') b|=8;
      if(cc=='v') b|=16;
      break;
    case 1:	/* a,s,d,f,g */
      if(cc=='a') b|=1;
      if(cc=='s') b|=2;
      if(cc=='d') b|=4;
      if(cc=='f') b|=8;
      if(cc=='g') b|=16;
      break;
    case 2:	/* q,w,e,r,t */
      if(cc=='q') b|=1;
      if(cc=='w') b|=2;
      if(cc=='e') b|=4;
      if(cc=='r') b|=8;
      if(cc=='t') b|=16;
      break;
    case 3:	/* 1,2,3,4,5 */
      if(c=='1' || c=='!') b|=1;
      if(c=='2' || c=='@') b|=2;
      if(c=='3' || c=='#') b|=4;
      if(c=='4' || c=='$') b|=8;
      if(c=='5' || c=='%') b|=16;
      break;

    /* right-hand side */
    case 4:	/* 0,9,8,7,6 */
      if(c=='0' || c==')') b|=1;
      if(c=='9' || c=='(') b|=2;
      if(c=='8' || c=='*') b|=4;
      if(c=='7' || c=='&') b|=8;
      if(c=='6' || c=='^') b|=16;
      break;
    case 5:	/* p,o,i,u,y */
      if(cc=='p') b|=1;
      if(cc=='o') b|=2;
      if(cc=='i') b|=4;
      if(cc=='u') b|=8;
      if(cc=='y') b|=16;
      break;
    case 6:	/* ent,l,k,j,h */
      if(c==10 || c=='\'') b|=1;
      if(cc=='l') b|=2;
      if(cc=='k') b|=4;
      if(cc=='j') b|=8;
      if(cc=='h') b|=16;
      break;
    case 7:	/* spc,dot,m,n,b */
      if(c==' ' || c=='`') b|=1;
      if(c=='.' || c=='>') b|=2;
      if(cc=='m') b|=4;
      if(cc=='n') b|=8;
      if(cc=='b') b|=16;
      break;
    }
  keyports[y]=((b^31)|0xe0);	/* some things need top bits to be 1 */
  }

}



int main(int argc,char *argv[])
{
sf = stdout;
scrn_freq = 25;
refresh_screen=0;

parseoptions(argc,argv);

initmem();
loadhelp();
zxpopen();

#ifdef OSS_SOUND_SUPPORT
if(sound)
  sound_init();
#endif
startsigsandtimer();
initts();
printf("Booting...\n");
mainloop();

/* not reached, but just in case */
exit_program();
return(0);	/* make -Wall happy */
}
